from loguru import logger
from ..data_source import DataSource
from ..result import Result, Results
from ..executor import Executor


class MysqlExecutor(Executor):

    def __init__(self, data_source: DataSource):
        super().__init__(data_source)

    def find(self, sql, args, bean_type, **options) -> Result:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        sql = sql.replace('?', '%s')
        connection = self.data_source.get_connection()
        connection.begin()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Result.empty()

            row = cursor.fetchone()
            if row is None:
                return Result.empty()

            return Result(row=row, bean_type=bean_type)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.commit()
            connection.close()

    def find_list(self, sql, args, bean_type, **options) -> Results:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        sql = sql.replace('?', '%s')
        connection = self.data_source.get_connection()
        connection.begin()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Results.empty()

            rows = cursor.fetchall()
            if rows is None:
                return Results.empty()

            return Results(rows=rows, bean_type=bean_type)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.commit()
            connection.close()

    def count(self, sql, args):
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        sql = sql.replace('?', '%s')
        connection = self.data_source.get_connection()
        connection.begin()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return None

            row = cursor.fetchone()
            if row is None:
                return None

            return row['COUNT(1)']
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.commit()
            connection.close()

    def update(self, sql, args):
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        sql = sql.replace('?', '%s')
        connection = self.data_source.get_connection()
        connection.begin()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return None
            return result
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.commit()
            connection.close()

    def insert(self, sql, args):
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        sql = sql.replace('?', '%s')
        connection = self.data_source.get_connection()
        connection.begin()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return None
            return result
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.commit()
            connection.close()

    def insert_bulk(self, sql, args):
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        sql = sql.replace('?', '%s')
        connection = self.data_source.get_connection()
        connection.begin()
        cursor = connection.cursor()
        try:
            for args in args:
                cursor.execute(sql, args)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.commit()
            connection.close()

    def raw(self, sql, args=()) -> Results:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        sql = sql.replace('?', '%s')
        connection = self.data_source.get_connection()
        connection.begin()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Results.empty()

            rows = cursor.fetchall()
            if rows is None:
                return Results.empty()

            return Results(rows=rows)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.commit()
            connection.close()



from typing import Any, List

from .wrapper import Wrapper
from .data_source import DataSource
from .sql_builder import build_select, build_count
from .result import Result, Results
from dataclasses import fields


class QueryWrapper(Wrapper):
    def __init__(self,
                 table: str,
                 data_source: DataSource,
                 tenant_field=None,
                 tenant_value=None,
                 logic_delete_field=None,
                 logic_delete_true=None,
                 logic_delete_false=None):
        super().__init__(tenant_field=tenant_field,
                         tenant_value=tenant_value,
                         logic_delete_field=logic_delete_field,
                         logic_delete_true=logic_delete_true,
                         logic_delete_false=logic_delete_false)
        self.table = table
        self.data_source = data_source
        self.result_type = data_source.get_data_structure(table)
        self.limit_ = None
        self.offset = None
        self.order_by = None
        self.field_list = []
        self.ignore_fields = []

    def select(self, *field_list) -> 'QueryWrapper':
        self.field_list = field_list
        return self

    def ignore(self, *field_list) -> 'QueryWrapper':
        self.ignore_fields = field_list
        return self

    def sort(self, *order_by) -> 'QueryWrapper':
        self.order_by = order_by
        return self

    def limit(self, limit: int) -> 'QueryWrapper':
        self.limit_ = limit
        return self

    def offset(self, offset: int) -> 'QueryWrapper':
        self.offset = offset
        return self

    def find(self, as_dict=False, **options) -> Any:
        self.handle_public_fields(**options)

        sql, args = self._build_select()
        result: Result = self.data_source.get_executor().find(sql, args, self.result_type, **options)
        if as_dict:
            return result.as_dict()
        return result.get()

    def find_list(self, as_dict=False, **options) -> List[Any]:
        self.handle_public_fields(**options)

        sql, args = self._build_select()
        results: Results = self.data_source.get_executor().find_list(sql, args, self.result_type, **options)
        if as_dict:
            return results.as_dict()
        return results.get()

    def find_page(self, page: int, page_size: int, as_dict=False, **options) -> (List[Any], int):
        self.handle_public_fields(**options)

        self.limit_ = page_size
        self.offset = (page - 1) * page_size
        sql, args = self._build_select()
        results: Results = self.data_source.get_executor().find_list(sql, args, self.result_type, **options)
        count = self.count()
        if as_dict:
            return results.as_dict(), count
        return results.get(), count

    def count(self):
        sql, args = build_count(self)
        return self.data_source.get_executor().count(sql, args)

    def _build_select(self) -> (str, tuple):
        if len(self.field_list) == 0:
            self.field_list = [field.name for field in fields(self.result_type) if field.name not in self.ignore_fields]
        return build_select(self)



import datetime

from .wrapper import Wrapper
from .sql_builder import build_update, build_delete
from dataclasses import fields


class UpdateWrapper(Wrapper):

    def __init__(self,
                 table,
                 data_source,
                 tenant_field=None,
                 tenant_value=None,
                 update_by_field=None,
                 update_at_field=None,
                 logic_delete_field=None,
                 logic_delete_true=None,
                 logic_delete_false=None):
        super().__init__(tenant_field=tenant_field,
                         tenant_value=tenant_value,
                         logic_delete_field=logic_delete_field,
                         logic_delete_true=logic_delete_true,
                         logic_delete_false=logic_delete_false)
        self.update_by_field = update_by_field
        self.update_at_field = update_at_field
        self.table = table
        self.data_source = data_source
        self.update_fields = {}

    def set(self, field, value) -> 'UpdateWrapper':
        self.update_fields[field] = value
        return self

    def read(self, entity) -> 'UpdateWrapper':
        if isinstance(entity, dict):
            for field in fields(self.data_source.get_data_structure(self.table)):
                if field.name != 'id' and field.name in entity:
                    self.update_fields[field.name] = entity[field.name]
            return self

        for field in fields(entity):
            if field.name != 'id':
                continue
            self.update_fields[field.name] = getattr(entity, field.name)
        return self

    def update(self, **options):
        if len(self.condition_tree.conditions) == 0:
            raise ValueError('不支持全量更新')

        self.handle_public_fields(**options)

        sql, args = build_update(self)
        return self.data_source.get_executor().update(sql, args)

    def delete(self, **options):
        if len(self.condition_tree.conditions) == 0:
            raise ValueError('不支持全量更新')

        self.handle_public_fields(**options)

        if options.get('logical_delete', False):
            self.set(options['logical_delete'], 1)
            return self.update(**options)
        else:
            sql, args = build_delete(self)
            return self.data_source.get_executor().update(sql, args)

    def handle_update_fields(self, **options) -> 'UpdateWrapper':
        if self.update_at_field is not None:
            self.set(self.update_at_field, datetime.datetime.now())
        if self.update_by_field is not None:
            self.set(self.update_by_field, options.get('update_by', 'system'))
        return self
